"""Test base.py"""


def test_cosmo_import():
    """Just test that we can import cosmo (which triggers many object constructions)"""
    from cosmograph import cosmo

    assert cosmo


def test_cosmo_graph_01_quests():
    # Specify some data
    data = {
        "Quest_Title": ["Dragon Hunt", "Mystic Voyage", "Treasure Seekers"],
        "Map_X": [120, 450, 100],
        "Map_Y": [75, 320, 210],
        "Aura_Color": ["Crimson", "Azure", "Emerald"],
        "Gear_Size": [15, 20, 10],
    }

    # Create a DataFrame containing the data
    import pandas as pd

    df = pd.DataFrame(data)

    # Make the graph

    from cosmograph import cosmo

    graph = cosmo(
        df,
        point_x_by="Map_X",
        point_y_by="Map_Y",
        point_color_by="Aura_Color",
        point_size_by="Gear_Size",
        point_size_scale=7,
        point_label_by="Quest_Title",
    )

    # Check that the graph is a widget

    import anywidget

    assert isinstance(
        graph, anywidget.widget.AnyWidget
    ), "'quest' graph wasn't an anywidget widget instance"

# 👇 Commented out because a simple cosmo does not support ingress functionality
# def test_cosmo_partial_and_ingress_01():
#     import pytest
#     import pandas as pd
#     from cosmograph import cosmo, Cosmograph

#     # Define the data with whimsical column names
#     data_potions = {
#         "potion_name": [
#             "Elixir of Vitality",
#             "Draught of Invisibility",
#             "Brew of Fireball",
#         ],
#         "Brewmaster": ["Elder Willow", "Mistress Shadow", "Master Ignis"],
#         "Ingredients": [
#             "Dragon Scale, Phoenix Feather, Unicorn Tears",
#             "Nightshade, Moonflower, Shadow Essence",
#             "Fireroot, Lava Crystal, Unicorn Tears, Ember Dust",
#         ],
#         "Location_X": [500, 620, 450],
#         "Location_Y": [300, 400, 350],
#         "potion_color": ["Blue", "White", "Red"],
#         "potion_difficulty": [25, 30, 20],
#         "potion_popularity": [1, 2, 3],
#     }

#     # Create a DataFrame to contain this data
#     df_potions = pd.DataFrame(data_potions)

#     # Test that Cosmograph widget works
#     asis = cosmo(
#         df_potions,
#         point_x_by="Location_X",
#         point_y_by="Location_Y",
#         point_label_by="potion_name",
#         point_size_by="potion_difficulty",
#         point_color_by="potion_color",
#     )
#     assert isinstance(asis, Cosmograph), "asis is not an instance of Cosmograph!"

#     # If you were to iterate over different datas or configurations, you could make
#     # a "partial" widget constructor that sets the common parameters (just don't include any data, points, or links arguments)
#     my_cosmo = cosmo(
#         point_x_by="Location_X",
#         point_y_by="Location_Y",
#         point_label_by="potion_name",
#         point_size_by="potion_difficulty",
#         point_color_by="potion_color",
#     )

#     # You my_cosmo is not a widget now...
#     assert not isinstance(my_cosmo, Cosmograph)
#     # ... but a callable that can be used to create widgets
#     assert callable(my_cosmo), "my_cosmo is not callable!"
#     # See, let's use it to create a widget
#     widget_from_my_cosmo = my_cosmo(df_potions)
#     assert isinstance(
#         widget_from_my_cosmo, Cosmograph
#     ), "widget_from_my_cosmo is not an instance of Cosmograph!"

#     # Use pytest.raises to check that TraitError is raised

#     from traitlets.traitlets import TraitError

#     with pytest.raises(
#         TraitError,
#         match="The 'point_color_by' trait of a Cosmograph instance expected a unicode string, not the function '<lambda>'",
#     ):
#         with_functional_color_failing = my_cosmo(
#             df_potions,
#             point_color_by=lambda x: int(
#                 "Unicorn Tears" in x["Ingredients"]
#             ),  # Invalid point_color_by: It's supposed to be a column name, but it's a function instead!!
#         )

#     # Introducing ingress. Ingress is a function, or sequence of functions, that
#     # transforms the input arguments before they are passed to the Cosmograph constructor.
#     #
#     # For instance, you could use it to specify that if point_color_by is a function,
#     # it should apply this function to the DataFrame rows, creating a temp column to
#     # accumulate the results, and then change the point_color_by argument to be the name
#     # of that column, thereby being compliant with the expected types.

#     def handle_functional_color(kwargs):
#         if callable(kwargs.get("point_color_by")):
#             point_color_by_func = kwargs["point_color_by"]
#             points_copy = kwargs["points"].copy()
#             points_copy["_point_color_by"] = points_copy.apply(
#                 point_color_by_func, axis=1
#             )
#             kwargs["point_color_by"] = "_point_color_by"
#             kwargs["points"] = points_copy
#         return kwargs

#     # Test case for "with_functional_color"
#     with_functional_color = my_cosmo(
#         df_potions,
#         ingress=handle_functional_color,
#         point_color_by=lambda x: int("Unicorn Tears" in x["Ingredients"]),
#     )
#     assert isinstance(
#         with_functional_color, Cosmograph
#     ), "with_functional_color is not an instance of Cosmograph!"



# import importlib.metadata
import pathlib
import json

import anywidget
from traitlets import Bool, Float, List, Int, Unicode, Union, Bytes, Any, observe, Dict

from .utils import get_buffered_arrow_table
from .export_project import export_project
from cosmograph.config import register_instance, unregister_instance, get_api_key

meta_path = pathlib.Path(__file__).parent / "static" / "meta.json"

try:
    with open(meta_path) as f:
        meta_data = json.load(f)
    js_file = meta_data["js"]
    css_file = meta_data["css"]
except (FileNotFoundError, json.JSONDecodeError, KeyError):
    js_file = "widget.js"
    css_file = "widget.css"
# with open(meta_path) as f:
#     meta_data = json.load(f)


class Cosmograph(anywidget.AnyWidget):
    _esm = pathlib.Path(__file__).parent / "static" / js_file
    _css = pathlib.Path(__file__).parent / "static" / css_file

    # Configuration parameters for Cosmograph
    # List of all configuration parameters that JS side support can be found in ./js/config-props.ts

    # Parameters based on parameters from Cosmos (https://github.com/cosmograph-org/cosmos/blob/main/src/config.ts)
    disable_simulation = Bool(None, allow_none=True).tag(sync=True)

    simulation_decay = Float(None, allow_none=True).tag(sync=True)
    simulation_gravity = Float(None, allow_none=True).tag(sync=True)
    simulation_center = Float(None, allow_none=True).tag(sync=True)
    simulation_repulsion = Float(None, allow_none=True).tag(sync=True)
    simulation_repulsion_theta = Float(None, allow_none=True).tag(sync=True)
    simulation_repulsion_quadtree_levels = Float(None, allow_none=True).tag(sync=True)
    simulation_link_spring = Float(None, allow_none=True).tag(sync=True)
    simulation_link_distance = Float(None, allow_none=True).tag(sync=True)
    simulation_link_dist_random_variation_range = List(
        None, default_value=None, allow_none=True
    ).tag(sync=True)
    simulation_repulsion_from_mouse = Float(None, allow_none=True).tag(sync=True)
    simulation_friction = Float(None, allow_none=True).tag(sync=True)
    simulation_cluster_strength = Float(None, allow_none=True).tag(sync=True)

    background_color = Union(
        [Unicode(None, allow_none=True), List(Float, allow_none=True)]
    ).tag(sync=True)
    space_size = Int(None, allow_none=True).tag(sync=True)
    point_color = Union(
        [Unicode(None, allow_none=True), List(Float, allow_none=True)]
    ).tag(sync=True)
    point_greyout_opacity = Float(None, allow_none=True).tag(sync=True)
    point_size = Float(None, allow_none=True).tag(sync=True)
    point_size_scale = Float(None, allow_none=True).tag(sync=True)
    hovered_point_cursor = Unicode(None, allow_none=True).tag(sync=True)
    render_hovered_point_ring = Bool(None, allow_none=True).tag(sync=True)
    hovered_point_ring_color = Union(
        [Unicode(None, allow_none=True), List(Float, allow_none=True)]
    ).tag(sync=True)
    focused_point_ring_color = Union(
        [Unicode(None, allow_none=True), List(Float, allow_none=True)]
    ).tag(sync=True)
    focused_point_index = Int(None, allow_none=True).tag(sync=True)
    show_focused_point_label = Bool(None, allow_none=True).tag(sync=True)
    render_links = Bool(None, allow_none=True).tag(sync=True)
    link_color = Union(
        [Unicode(None, allow_none=True), List(Float, allow_none=True)]
    ).tag(sync=True)
    link_greyout_opacity = Float(None, allow_none=True).tag(sync=True)
    link_width = Float(None, allow_none=True).tag(sync=True)
    link_width_scale = Float(None, allow_none=True).tag(sync=True)
    curved_links = Bool(None, allow_none=True).tag(sync=True)
    curved_link_segments = Int(None, allow_none=True).tag(sync=True)
    curved_link_weight = Float(None, allow_none=True).tag(sync=True)
    curved_link_control_point_distance = Float(None, allow_none=True).tag(sync=True)
    link_arrows = Bool(None, allow_none=True).tag(sync=True)
    link_arrows_size_scale = Float(None, allow_none=True).tag(sync=True)
    link_visibility_distance_range = List(
        Float, default_value=None, allow_none=True
    ).tag(sync=True)
    link_visibility_min_transparency = Float(None, allow_none=True).tag(sync=True)
    use_quadtree = Bool(None, allow_none=True).tag(sync=True)
    show_FPS_monitor = Bool(None, allow_none=True).tag(sync=True)
    pixel_ratio = Float(None, allow_none=True).tag(sync=True)
    # Set to True by default until it becomes the default in Cosmograph library
    scale_points_on_zoom = Bool(True, allow_none=True).tag(sync=True)
    scale_links_on_zoom = Bool(None, allow_none=True).tag(sync=True)
    initial_zoom_level = Float(None, allow_none=True).tag(sync=True)
    disable_zoom = Bool(None, allow_none=True).tag(sync=True)
    enable_drag = Bool(None, allow_none=True).tag(sync=True)
    fit_view_on_init = Bool(None, allow_none=True).tag(sync=True)
    fit_view_delay = Float(None, allow_none=True).tag(sync=True)
    fit_view_padding = Float(None, allow_none=True).tag(sync=True)
    fit_view_duration = Float(None, allow_none=True).tag(sync=True)
    fit_view_by_points_in_rect = List(
        List(Float), default_value=None, allow_none=True
    ).tag(sync=True)
    random_seed = Union(
        [Int(None, allow_none=True), Unicode(None, allow_none=True)]
    ).tag(sync=True)
    point_sampling_distance = Int(None, allow_none=True).tag(sync=True)

    # Polygonal selection configuration
    polygonal_selector_stroke_color = Unicode(None, allow_none=True).tag(sync=True)
    polygonal_selector_line_width = Float(None, allow_none=True).tag(sync=True)

    # Parameters based on parameters from Cosmograph library
    point_id_by = Unicode(None, allow_none=True).tag(sync=True)
    point_index_by = Unicode(None, allow_none=True).tag(sync=True)

    point_color_by = Unicode(None, allow_none=True).tag(sync=True)
    point_color_palette = List(Unicode, default_value=None, allow_none=True).tag(
        sync=True
    )
    point_color_by_map = Dict(
        key_trait=Unicode(), value_trait=Union([Unicode(), List(Float())]), default_value=None, allow_none=True
    ).tag(sync=True)
    point_color_strategy = Unicode(None, allow_none=True).tag(sync=True)

    point_size_by = Unicode(None, allow_none=True).tag(sync=True)
    point_size_range = List(Float, default_value=None, allow_none=True).tag(sync=True)
    point_size_strategy = Unicode(None, allow_none=True).tag(sync=True)
    point_label_by = Unicode(None, allow_none=True).tag(sync=True)
    point_label_weight_by = Unicode(None, allow_none=True).tag(sync=True)
    point_x_by = Unicode(None, allow_none=True).tag(sync=True)
    point_y_by = Unicode(None, allow_none=True).tag(sync=True)
    point_cluster_by = Unicode(None, allow_none=True).tag(sync=True)
    point_cluster_strength_by = Unicode(None, allow_none=True).tag(sync=True)
    cluster_positions_map = Dict(
        key_trait=Unicode(), value_trait=List(Float()), default_value=None, allow_none=True
    ).tag(sync=True)
    point_include_columns = List(Unicode, default_value=None, allow_none=True).tag(
        sync=True
    )
    point_timeline_by = Unicode(None, allow_none=True).tag(sync=True)

    link_source_by = Unicode(None, allow_none=True).tag(sync=True)
    link_source_index_by = Unicode(None, allow_none=True).tag(sync=True)
    link_target_by = Union([Unicode(), List(Unicode())], default_value=None, allow_none=True).tag(sync=True)
    link_target_index_by = Unicode(None, allow_none=True).tag(sync=True)
    link_color_by = Unicode(None, allow_none=True).tag(sync=True)
    link_width_by = Unicode(None, allow_none=True).tag(sync=True)
    link_arrow_by = Unicode(None, allow_none=True).tag(sync=True)
    link_strength_by = Unicode(None, allow_none=True).tag(sync=True)
    link_strength_range = List(Float, default_value=None, allow_none=True).tag(
        sync=True
    )
    link_include_columns = List(Unicode, default_value=None, allow_none=True).tag(
        sync=True
    )
    link_timeline_by = Unicode(None, allow_none=True).tag(sync=True)

    show_labels = Bool(None, allow_none=True).tag(sync=True)
    show_dynamic_labels = Bool(None, allow_none=True).tag(sync=True)
    show_labels_for = List(Unicode, default_value=None, allow_none=True).tag(sync=True)
    show_top_labels = Bool(None, allow_none=True).tag(sync=True)
    show_top_labels_limit = Int(None, allow_none=True).tag(sync=True)
    static_label_weight = Float(None, allow_none=True).tag(sync=True)
    dynamic_label_weight = Float(None, allow_none=True).tag(sync=True)
    label_margin = Float(None, allow_none=True).tag(sync=True)
    show_cluster_labels = Bool(None, allow_none=True).tag(sync=True)
    point_label_color = Unicode(None, allow_none=True).tag(sync=True)
    point_label_font_size = Float(None, allow_none=True).tag(sync=True)
    cluster_label_font_size = Float(None, allow_none=True).tag(sync=True)
    scale_cluster_labels = Bool(None, allow_none=True).tag(sync=True)

    label_padding = List(Float, default_value=None, allow_none=True).tag(sync=True)

    show_hovered_point_label = Bool(None, allow_none=True).tag(sync=True)
    use_point_color_strategy_for_cluster_labels = Bool(None, allow_none=True).tag(sync=True)
    select_cluster_on_label_click = Bool(None, allow_none=True).tag(sync=True)
    select_point_on_click = Union(
        [Bool(None, allow_none=True), Unicode(None, allow_none=True)]
    ).tag(sync=True)
    select_point_on_label_click = Union(
        [Bool(None, allow_none=True), Unicode(None, allow_none=True)]
    ).tag(sync=True)
    focus_point_on_click = Bool(None, allow_none=True).tag(sync=True)
    focus_point_on_label_click = Bool(None, allow_none=True).tag(sync=True)
    components_display_state_mode = Union(
        [Unicode(None, allow_none=True), Bool(None, allow_none=True)]
    ).tag(sync=True)

    # Advanced configuration options
    status_indicator_mode = Union(
        [Unicode(None, allow_none=True), Bool(None, allow_none=True)]
    ).tag(sync=True)
    preserve_point_positions_on_data_update = Bool(None, allow_none=True).tag(sync=True)

    # Not related to Cosmograph configuration settings
    disable_point_size_legend = Bool(None, allow_none=True).tag(sync=True)
    disable_link_width_legend = Bool(None, allow_none=True).tag(sync=True)
    disable_point_color_legend = Bool(None, allow_none=True).tag(sync=True)
    disable_link_color_legend = Bool(None, allow_none=True).tag(sync=True)

    # Points and links are Pandas DataFrames that will be passed
    # to the JS side widget as an IPC (Inter-Process Communication) stream
    _ipc_points = Bytes(None, allow_none=True).tag(sync=True)
    points = Any()

    _ipc_links = Bytes(None, allow_none=True).tag(sync=True)
    links = Any()

    # The following are used to store values from JS side widget
    clicked_point_index = Int(None, allow_none=True).tag(sync=True)
    clicked_point_id = Unicode(None, allow_none=True).tag(sync=True)
    clicked_cluster = Any(None, allow_none=True).tag(sync=True)
    selected_point_indices = List(Int, allow_none=True).tag(sync=True)
    selected_point_ids = List(Unicode, allow_none=True).tag(sync=True)
    cosmograph_config = Dict(default_value={}, allow_none=True).tag(sync=True)

    api_key = Unicode(None, allow_none=True)

    @observe("points")
    def changePoints(self, change):
        points = change.new
        self._ipc_points = get_buffered_arrow_table(points)

    @observe("links")
    def changeLinks(self, change):
        links = change.new
        self._ipc_links = get_buffered_arrow_table(links)

    # Public cosmograph widget methods that can be called from Python
    # to interact with the Cosmograph 👇
    def select_point_by_index(self, index):
        self.send({"type": "select_point_by_index", "index": index})

    def select_point_by_id(self, id):
        self.send({"type": "select_point_by_id", "id": id})

    def select_points_by_indices(self, indices):
        self.send({"type": "select_points_by_indices", "indices": indices})

    def select_points_by_ids(self, ids):
        self.send({"type": "select_points_by_ids", "ids": ids})

    def activate_rect_selection(self):
        self.send({"type": "activate_rect_selection"})

    def deactivate_rect_selection(self):
        self.send({"type": "deactivate_rect_selection"})

    def activate_polygonal_selection(self):
        self.send({"type": "activate_polygonal_selection"})

    def deactivate_polygonal_selection(self):
        self.send({"type": "deactivate_polygonal_selection"})

    def select_points_in_polygon(self, polygon):
        if not isinstance(polygon, list) or len(polygon) < 3:
            raise ValueError("Polygon must be a list of at least 3 coordinate pairs")
        for point in polygon:
            if not isinstance(point, (list, tuple)) or len(point) != 2:
                raise ValueError("Each polygon point must be a coordinate pair [x, y]")
        self.send({"type": "select_points_in_polygon", "polygon": polygon})

    def unselect_points_by_indices(self, indices):
        self.send({"type": "unselect_points_by_indices", "indices": indices})

    def fit_view(self):
        self.send({"type": "fit_view"})

    def fit_view_by_indices(self, indices, duration=None, padding=None):
        self.send(
            {
                "type": "fit_view_by_indices",
                "indices": indices,
                "duration": duration,
                "padding": padding,
            }
        )

    def fit_view_by_ids(self, ids, duration=None, padding=None):
        self.send(
            {
                "type": "fit_view_by_ids",
                "ids": ids,
                "duration": duration,
                "padding": padding,
            }
        )

    def fit_view_by_coordinates(self, coordinates, duration=None, padding=None):
        self.send(
            {
                "type": "fit_view_by_coordinates",
                "coordinates": coordinates,
                "duration": duration,
                "padding": padding,
            }
        )

    def focus_point_by_index(self, index=None):
        self.send({"type": "focus_point_by_index", "index": index})

    def focus_point(self, id=None):
        self.send({"type": "focus_point", "id": id})

    def start(self, alpha=None):
        self.send({"type": "start", "alpha": alpha})

    def pause(self):
        self.send({"type": "pause"})

    def restart(self):
        self.send({"type": "restart"})

    def step(self):
        self.send({"type": "step"})

    def capture_screenshot(self):
        self.send({"type": "capture_screenshot"})

    def export_project_by_name(self, project_name: str):
        if not self.api_key:
            raise ValueError("API key is required for project export")
        if not project_name or not project_name.strip():
            raise ValueError("Project name is required and cannot be empty")
        try:
            return export_project(self.api_key, project_name, self.points, self.links, self.cosmograph_config)
        except Exception as e:
            raise RuntimeError(f"Failed to export project '{project_name}': {str(e)}") from e

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # Register this instance to receive API key updates
        try:
            register_instance(self)
        except Exception as e:
            # Log warning but don't fail initialization
            print(f"Warning: Failed to register instance for API key updates: {e}")

        # If no API key was provided but global one exists, use it
        if self.api_key is None:
            global_api_key = get_api_key()
            if global_api_key is not None:
                self.api_key = global_api_key

    def __del__(self):
        # Clean up by unregistering when the instance is deleted
        try:
            unregister_instance(self)
        except Exception:
            # Ignore errors during cleanup
            pass
        # Only call super().__del__() if it exists
        if hasattr(super(), '__del__'):
            super().__del__()



"""
Utils to prepare data for cosmos
"""

import json
import inspect
from typing import Dict, Any
from functools import partial
import re

from IPython.display import HTML, Javascript, display as ipython_display
from dol import (
    TextFiles,
    Files,
    JsonFiles,
    wrap_kvs,
    filt_iter,
    invertible_maps,
    add_ipython_key_completions,
    Pipe,
)
from dol.sources import AttrContainer

import pandas as pd

# --------------------------------------------------------------------------------------
# type annotations

CosmoKwargs = Dict[str, Any]

# --------------------------------------------------------------------------------------
# Constants and data access

try:
    import importlib.resources

    _files = importlib.resources.files  # only valid in 3.9+
except AttributeError:
    import importlib_resources  # needs pip install

    _files = importlib_resources.files

files = _files("cosmograph")
data_dir = files / "data"
data_dir_path = str(data_dir)
js_dir = files / "js"
js_dir_path = str(js_dir)

data_files = Files(data_dir_path)
json_files = filt_iter.suffixes(".json")(JsonFiles(data_dir_path))


# color_names_set = set(json_files['color_names.json'])  # removed because problematic on windows

# --------------------------------------------------------------------------------------
# Pipeline for Cosmo


class Pipeline(Pipe):
    def add(self, *additional_funcs):
        funcs = tuple(self.funcs) + tuple(additional_funcs)
        return Pipeline(*funcs)


# --------------------------------------------------------------------------------------
# Extracting the interface from the data

import typing
from typing import Any, Callable
from i2 import Sig, Param, params_to_docstring

PARAMS_SSOT_PATH = data_dir / "params_ssot.json"

# TODO: add ssot validation


def _params_ssot(param_names=None):
    params_ssot = json.loads(PARAMS_SSOT_PATH.read_text())
    if param_names is not None:
        params_ssot = [d for d in params_ssot if d["name"] in param_names]
    return params_ssot


def cosmograph_base_signature(param_names=None):
    params_ssot = _params_ssot(param_names)
    for d in params_ssot:
        d.pop("description", None)  # delete description if any
        d.update(
            annotation=str_to_annotation(d["annotation"])
        )  # convert annotation to type
        d["kind"] = Sig.KEYWORD_ONLY

    return Sig([Param(**d) for d in params_ssot])


def validate_signature(sig):
    """Validate a signature.
    Namely, check that all non-None default types are "subclasses" of the annotation type.

    TODO: Implement
    """
    pass


def cosmograph_base_docs(param_names=None, take_name_of_types=True):
    """Get the params information part of a docstring"""
    params_ssot = _params_ssot(param_names)
    return params_to_docstring(params_ssot)


def add_cosmo_param_descriptions(func, param_names=None):
    """
    Decorator to add parameter descriptions from SSOT data to cosmo function.

    Simply appends SSOT parameter descriptions to the existing Args section.

    Args:
        func: The cosmo function to decorate.
        param_names: Optional list of param names to filter SSOT (uses _params_ssot).
    """
    params_ssot = _params_ssot(param_names)
    param_docs = {p["name"]: p for p in params_ssot}

    sig = inspect.signature(func)
    relevant_params = [
        p for name, p in param_docs.items()
        if name in sig.parameters and p["description"]
    ]

    if relevant_params:
        # Simply append SSOT parameters to the existing docstring
        ssot_docs = "\n".join(f"        {p['name']}: {p['description']}" for p in relevant_params)
        func.__doc__ = func.__doc__.rstrip() + "\n" + ssot_docs + "\n"

    return func


def is_parameterized_type(obj):
    """
    Determines if an object is a parameterized type (e.g., list[int]) or not.

    Args:
        obj: The object to check.

    Returns:
        True if it's a parameterized type, False otherwise.
    """
    return typing.get_origin(obj) is not None


def annotation_to_str(annotation, *, remove_typing_prefix=True):
    """
    Encodes Python type annotations as strings for JSON serialization.

    This is part of a specialized Python-to-string codec designed to encode and decode
    Python type annotations in a way that allows them to be serialized and deserialized
    in JSON-compatible formats.

    Args:
        annotation: The type annotation to encode.

    Returns:
        A string representation of the annotation.

    Raises:
        ValueError: If the annotation type is unknown.

    Examples:
    >>> original = [
    ...     bool,
    ...     float,
    ...     list[Any],
    ...     typing.Union[str, list[float]],
    ...     int,
    ...     str,
    ...     list[float],
    ...     list[list[float]],
    ...     typing.Union[int, str],
    ...     list[str],
    ...     object,
    ...     list[int],
    ...     typing.Callable[[typing.Dict[str, Any]], Any]
    ... ]
    >>> encoded = list(map(annotation_to_str, original))
    >>> encoded  # doctest: +NORMALIZE_WHITESPACE
    ['bool', 'float', 'list[Any]', 'Union[str, list[float]]', 'int', 'str', \
    'list[float]', 'list[list[float]]', 'Union[int, str]', 'list[str]', 'object', 'list[int]', \
    'Callable[[Dict[str, Any]], Any]']
    """
    if isinstance(annotation, str):
        annotation_str = annotation
    elif annotation == Sig.empty:
        annotation_str = "Any"
    elif is_parameterized_type(annotation):
        annotation_str = str(annotation)
    elif isinstance(annotation, type):
        annotation_str = annotation.__name__
    else:
        raise ValueError(f"Unknown annotation type: {annotation}")

    if remove_typing_prefix:
        typing_module_pattern = r"(?:(?<=\W)|^)typing\."
        annotation_str = re.sub(typing_module_pattern, "", annotation_str)

    return annotation_str


def default_is_safe(string: str) -> bool:
    """
    Checks if a string is safe to evaluate as a type annotation.

    The string is considered safe if:
    - It contains only alphanumericals, spaces, dots, commas, and brackets.
    - It starts with one of a predefined list of allowed prefixes.

    Args:
        string: The string to check.

    Returns:
        True if the string is safe, False otherwise.
    """
    allowed_prefixes = ["bool", "float", "list", "typing", "int", "object", "str"]
    allowed_chars = re.compile(r"^[\w\s.,\[\]()]+$")

    return allowed_chars.match(string) is not None and any(
        string.strip().startswith(prefix) for prefix in allowed_prefixes
    )


def str_to_annotation(string, is_safe: Callable[[str], bool] = default_is_safe):
    """
    Decodes a string representation of a Python type annotation back into the annotation.

    Args:
        string: The string representation of the annotation.
        is_safe: A function that determines if the string is safe to evaluate.

    Returns:
        The decoded Python type annotation.

    Raises:
        ValueError: If the string is not safe to evaluate.
    """
    if not is_safe(string):
        raise ValueError(
            f"The string '{string}' is not considered safe for evaluation."
        )
    return eval(string)


# --------------------------------------------------------------------------------------
# General/Misc utils



def move_to_front(df: pd.DataFrame, cols) -> pd.DataFrame:
    """
    Move the columns in `cols` to the front of the DataFrame
    """
    return df[cols + [col for col in df.columns if col not in cols]]


def ordered_unique(iterable):
    seen = set()
    seen_add = seen.add
    return (x for x in iterable if not (x in seen or seen_add(x)))


class IpythonObjects:
    def __init__(self, *objs):
        self.objs = objs

    def display(self):
        for obj in self.objs:
            return ipython_display(obj)

    # def __repr__(self):
    #     return self.display()


def add_attributes(obj, **attrs):
    for k, v in attrs.items():
        setattr(obj, k, v)
    return obj


def camel_to_snake_case(name):
    # Replace any non-word character with an underscore
    name = re.sub(r"\W+", "_", name)
    # Insert an underscore before any capital letter (except the first one)
    name = re.sub(r"([a-z])([A-Z])", r"\1_\2", name).lower()
    return name


def snake_to_camel_case(name, first_char_trans=str.lower):
    # Split the string into words separated by underscores
    words = name.split("_")
    # Convert each word to title case (i.e., with the first letter capitalized)
    words = [w.capitalize() for w in words]
    # Join the words together into a single string
    result = "".join(words)
    # process the first character (possibly)
    if first_char_trans and result:
        result = first_char_trans(result[0]) + result[1:]
    return result


def _assert_camel_and_snake_sanity(camel_cases, snake_cases):
    """
    Make sure that our camel to snake functions can actually fall back on their feet.
    (So we don't have to keep the mapping around)
    """
    # For now, they don't (need to discuss) so muting this function


#     agree_1 = [
#         x == y for x, y in zip(
#             map(camel_to_snake_case, camel_cases),
#             list(snake_cases)
#         )
#     ]
#     agree_2 = [
#         x == y for x, y in zip(
#             list(camel_cases),
#             map(snake_to_camel_case, snake_cases),
#         )
#     ]
#     agree = [x and y for x, y in zip(agree_1, agree_2)]
#     if not all(agree):
#         i = agree.index(False)
#         raise AssertionError(
#             f"{list(camel_cases)[i]} and {list(snake_cases)[i]} do not agree"
#         )

# _cosmos_config_info = cosmos_config_info()


# --------------------------------------------------------------------------------------
# Old stuff (TODO: Deprecate and remove)


def _postprocess(func, egress):
    return Pipe(func, egress)


def postprocess(egress):
    return partial(_postprocess, egress=egress)

display_output = postprocess(ipython_display)
to_html_obj = postprocess(HTML)
to_js_obj = postprocess(Javascript)


@add_ipython_key_completions
@wrap_kvs(key_of_id=lambda x: x[: -len(".js")], id_of_key=lambda x: x + ".js")
@filt_iter(filt=lambda x: x.endswith(".js"))
class JsFiles(TextFiles):
    """A store of js files"""


_replace_non_alphanumerics_by_underscore = partial(re.compile(r"\W").sub, "_")


# Note: js_files_as_attrs is not used in the module, but can be useful when working
# in a notebook, or console, where we might want the convenience of tab-completion of
# attributes
def js_files_as_attrs(rootdir):
    """
    Will make a JsFiles, but where the keys are available as attributes.
    To do so, any non alphanumerics of file name are replaced with underscore,
    and there can be no two files that collide with that key transformation!
    """
    s = JsFiles(rootdir)
    key_for_id = {id_: _replace_non_alphanumerics_by_underscore(id_) for id_ in s}
    key_for_id, id_for_key = invertible_maps(key_for_id)
    return AttrContainer(
        **wrap_kvs(s, key_of_id=key_for_id.get, id_of_key=id_for_key.get)
    )


def _nodes_from_links(links):
    def _yield_nodes_from_links(links):
        for link in links:
            yield link["source"]
            yield link["target"]

    return [{"id": x} for x in ordered_unique(_yield_nodes_from_links(links))]





"""Module for handling file uploads to Cosmograph server."""
from typing import Any

import requests
import json

from .config import API_BASE, logger


def upload_file(api_key: str, data: dict[str, Any], project_id: str) -> dict[str, Any]:
    """Generate signed URL and upload file to Cosmograph server.

    Args:
        api_key: Cosmograph API key
        data: Dictionary containing file data (file_name, content_length, content)
        project_id: Project ID to associate the file with

    Raises:
        ValueError: If file upload or URL generation fails

    Returns:
        Dictionary containing original data with added 'upload_url' key
        containing the generated upload URL
    """

    try:
        response = requests.post(
            f"{API_BASE}/publicApi.generateSignedUploadUrl",
            json={
                "json": {
                    "apiKey": api_key,
                    "projectId": project_id,
                    "fileName": data["file_name"],
                    "contentLength": data["content_length"],
                    "contentType": "application/parquet"
                },
            },
        )
        response.raise_for_status()
        # logger.info("Response: %s", json.dumps(response.json(), indent=4))
        response_json = response.json()
        try:
            upload_url = response_json["result"]["data"]["json"]["url"]
        except (KeyError, TypeError) as e:
            logger.error("❌ Unexpected response format: %s", response_json)
            msg = f"Failed to parse upload URL from response: {e}"
            raise ValueError(msg) from e
    except requests.RequestException as e:
        logger.error("❌ Failed to get upload URL: %s", e)
        msg = f"Failed to get upload URL: {e}"
        raise ValueError(msg) from e

    try:
        upload_response = requests.put(
            upload_url,
            data=data["content"],
            headers={"Content-Type": "application/parquet"}
        )
        upload_response.raise_for_status()
        logger.info("✅ File '%s' upload completed successfully", data["file_name"])
    except requests.RequestException as e:
        logger.error("❌ Failed to upload file '%s': %s", data["file_name"], e)
        msg = f"Failed to upload file: {e}"
        raise ValueError(msg) from e

    return {**data, "upload_url": upload_url}



"""Configuration settings and logger setup for the Cosmograph export module."""
import logging
import sys

API_BASE = "https://run.cosmograph.app/api/trpc"
logger = logging.getLogger(__name__)

logging.basicConfig(
  level=logging.INFO,
  format='%(levelname)s - %(message)s',
  stream=sys.stdout
)



"""Module for creating Cosmograph projects with uploaded data files."""

from typing import Any, Optional
import requests
import json

from .config import API_BASE, logger


def create_empty_project(
    api_key: str,
    project_name: str,
    description: str = "Project created via API",
) -> dict[str, Any]:
    """Create an empty Cosmograph project.

    Args:
        api_key: Cosmograph API key
        project_name: Name for the project
        description: Description for the project

    Returns:
        dict: API response from project creation containing project ID

    Raises:
        ValueError: If project creation fails
    """
    try:
        config_json = {
            "json": {
                "apiKey": api_key,
                "config": {
                    "title": project_name,
                    "description": description,
                },
            },
        }
        # logger.info("Creating empty project with config: %s", json.dumps(config_json, indent=4))
        response = requests.post(
            f"{API_BASE}/publicApi.createProject",
            json=config_json,
        )
        response.raise_for_status()

        result = response.json()
        # logger.info("✅ Empty project created successfully: %s", json.dumps(result, indent=4))
        return result
    except requests.RequestException as e:
        logger.error("❌ Failed to create empty project '%s': %s", project_name, e)
        msg = f"Failed to create empty project: {e}"
        raise ValueError(msg) from e


def create_project(
    api_key: str,
    project_name: str,
    points_data: Optional[dict[str, Any]] = None,
    links_data: Optional[dict[str, Any]] = None,
    cosmograph_config: Optional[dict[str, Any]] = None,
) -> dict[str, Any]:
    """Create a new Cosmograph project with the uploaded files.

    Raises:
        ValueError: If project creation fails

    Args:
        api_key: Cosmograph API key
        project_name: Name for the project
        points_data: Data for points table
        links_data: Data for links table
        cosmograph_config: Cosmograph config

    Returns:
        dict: API response from project creation

    """

    cosmograph_config = cosmograph_config or {}
    data_sources = []
    column_mapping = {
        "points": {"columns": {}},
        "links": {"columns": {}},
    }

    # Configure points data source and column mapping
    if points_data:
        data_sources.append({
            "type": "file",
            "tableName": "point_table",
            "fileName": points_data["file_name"],
        })

        # Map points columns according to API specifications
        points_columns = {}

        # Required columns
        point_id_by = cosmograph_config.get("pointIdBy")
        if point_id_by:
            points_columns["id"] = point_id_by

        # Optional columns - map only if present in config
        point_label_by = cosmograph_config.get("pointLabelBy")
        if point_label_by:
            points_columns["label"] = point_label_by

        point_color_by = cosmograph_config.get("pointColorBy")
        if point_color_by:
            points_columns["color"] = point_color_by

        point_size_by = cosmograph_config.get("pointSizeBy")
        if point_size_by:
            points_columns["size"] = point_size_by

        point_x_by = cosmograph_config.get("pointXBy")
        if point_x_by:
            points_columns["x"] = point_x_by

        point_y_by = cosmograph_config.get("pointYBy")
        if point_y_by:
            points_columns["y"] = point_y_by

        point_cluster_by = cosmograph_config.get("pointClusterBy")
        if point_cluster_by:
            points_columns["cluster"] = point_cluster_by

        point_timeline_by = cosmograph_config.get("pointTimelineBy")
        if point_timeline_by:
            points_columns["time"] = point_timeline_by

        if points_columns:
            column_mapping["points"] = {
                "columns": points_columns,
                "tableName": "point_table",
            }

    # Configure links data source and column mapping
    if links_data:
        data_sources.append({
            "type": "file",
            "tableName": "link_table",
            "fileName": links_data["file_name"],
        })

        # Map links columns according to API specifications
        links_columns = {}

        # Required columns
        link_source_by = cosmograph_config.get("linkSourceBy")
        link_target_by = cosmograph_config.get("linkTargetBy")
        if link_source_by and link_target_by:
            links_columns["source"] = link_source_by
            links_columns["target"] = link_target_by

            # Optional columns - map only if present in config
            link_width_by = cosmograph_config.get("linkWidthBy")
            if link_width_by:
                links_columns["width"] = link_width_by

            link_color_by = cosmograph_config.get("linkColorBy")
            if link_color_by:
                links_columns["color"] = link_color_by

            # Link timeline support
            link_timeline_by = cosmograph_config.get("linkTimelineBy")
            if link_timeline_by:
                links_columns["time"] = link_timeline_by

            column_mapping["links"] = {
                "columns": links_columns,
                "tableName": "link_table",
            }

    try:
        config_json = {
            "json": {
                "apiKey": api_key,
                "config": {
                    "title": project_name,
                    "dataSources": data_sources,
                    "columnMapping": column_mapping,
                    "cosmographConfig": cosmograph_config,
                },
            },
        }
        # logger.info("Config JSON: %s", json.dumps(config_json, indent=4))
        response = requests.post(
            f"{API_BASE}/publicApi.upsertProjectByName",
            json=config_json,
        )
        response.raise_for_status()

        return response.json()
    except requests.RequestException as e:
        logger.error("❌ Failed to create project '%s': %s", project_name, e)
        msg = f"Failed to create project: {e}"
        raise ValueError(msg) from e



"""Module for converting pandas DataFrames to Parquet format for Cosmograph uploads."""
from io import BytesIO
from typing import Any

import pandas as pd


def prepare_parquet_data(df: pd.DataFrame, filename: str) -> dict[str, Any]:
    """Convert DataFrame to Parquet format and prepare upload data.

    Args:
        df: Input DataFrame to convert
        filename: Name for the output file

    Returns:
        Dictionary containing file metadata and content:
        - file_name: Output filename
        - content: Parquet bytes content
        - content_length: Byte length of content

    """
    buffer = BytesIO()
    df.to_parquet(buffer)
    buffer.seek(0)
    content = buffer.read()
    content_length = len(content)

    return {
        "file_name": filename,
        "content": content,
        "content_length": content_length,
    }



"""Configuration settings for cosmograph."""

# Global API key variable
_GLOBAL_API_KEY = None
_COSMOGRAPH_INSTANCES = set()


def set_api_key(api_key: str) -> None:
    """
    Set a global API key to be used by all cosmograph instances.

    Args:
        api_key: The API key to use
    """
    global _GLOBAL_API_KEY
    _GLOBAL_API_KEY = api_key

    # Update API key for all existing instances
    for instance in _COSMOGRAPH_INSTANCES:
        instance.api_key = api_key


def get_api_key() -> str:
    """
    Get the current global API key.

    Returns:
        The current global API key, or None if not set
    """
    return _GLOBAL_API_KEY


def register_instance(instance) -> None:
    """
    Register a cosmograph instance to receive API key updates.

    Args:
        instance: A Cosmograph instance
    """
    _COSMOGRAPH_INSTANCES.add(instance)


def unregister_instance(instance) -> None:
    """
    Unregister a cosmograph instance.

    Args:
        instance: A Cosmograph instance
    """
    if instance in _COSMOGRAPH_INSTANCES:
        _COSMOGRAPH_INSTANCES.remove(instance)



"""Module for Cosmograph project export functionality."""

from typing import Any, Optional

import pandas as pd
import json

from .prepare_parquet import prepare_parquet_data
from .upload_file import upload_file
from .create_project import create_project, create_empty_project
from .config import logger


def export_project(
    api_key: Optional[str] = None,
    project_name: str = "Cosmograph Project",
    points: Optional[pd.DataFrame] = None,
    links: Optional[pd.DataFrame] = None,
    cosmograph_config: Optional[dict[str, Any]] = None,
) -> dict[str, Any]:
    """Export data to a Cosmograph project.

    Args:
        api_key: Cosmograph API key
        project_name: Name for the project
        points: DataFrame containing point data
        links: DataFrame containing link data
        cosmograph_config: Cosmograph configuration

    Returns:
        Dict containing the project ID

    Raises:
        ValueError: If no API key is provided.

    """  # noqa: E117
    if not api_key:
        error_msg = "No API key provided. Please set an API key using set_api_key() or pass it directly to the Cosmograph constructor."
        logger.error("❌ %s", error_msg)
        raise ValueError(error_msg)

    logger.info("🚀 Exporting project: %s (%s points, %s links)",
        project_name,
        len(points) if points is not None else 0,
        len(links) if links is not None else 0)

    # Step 1: Create empty project first
    try:
        project_result = create_empty_project(api_key, project_name)
        project_id = project_result["result"]["data"]["json"]["id"]
    except Exception as e:
        logger.error("❌ Failed to create empty project: %s", e)
        raise ValueError(f"Failed to create empty project: {e}") from e

    # Step 2: Upload files with project ID
    points_data = None
    links_data = None
    if points is not None and not points.empty:
        try:
            points_data = prepare_parquet_data(points, f"{project_name}_points.parquet")
            points_data = upload_file(api_key, points_data, project_id)
        except Exception as e:
            logger.error("❌ Failed to upload points data: %s", e)
            raise ValueError(f"Failed to process points data: {e}") from e

    if links is not None and not links.empty:
        try:
            links_data = prepare_parquet_data(links, f"{project_name}_links.parquet")
            links_data = upload_file(api_key, links_data, project_id)
        except Exception as e:
            logger.error("❌ Failed to upload links data: %s", e)
            raise ValueError(f"Failed to process links data: {e}") from e

    # Step 3: Update project with uploaded files and configuration
    if points_data or links_data or cosmograph_config:
        try:
            create_project(
                api_key=api_key,
                project_name=project_name,
                points_data=points_data,
                links_data=links_data,
                cosmograph_config=cosmograph_config,
            )
            # logger.info("Result: %s", json.dumps(result, indent=4))
        except Exception as e:
            logger.error("❌ Failed to update project: %s", e)
            raise ValueError(f"Failed to update project: {e}") from e

    project_url = f"https://run.cosmograph.app/project/{project_id}"
    logger.info("✅ Project ready: %s", project_url)

    return project_id



"""
Python binders to cosmograph functionality


>>> import pandas as pd
>>> from cosmograph import cosmo
>>>
>>> single_link = {
...     'points': [{'id': '0'}, {'id': '1'}],
...     'links': [{'source': '0', 'target': '1'}],
... }
>>> points, links = map(pd.DataFrame, single_link.values())
>>> g = cosmo(
...     points=points,
...     links=links,
...     point_id_by='id',
...     link_source_by='source',
...     link_target_by='target',
...     simulation_gravity=0,
...     simulation_center=1,
... )


"""

from cosmograph.base import cosmo, base_cosmo
from cosmograph.widget import Cosmograph
from cosmograph.config import set_api_key, get_api_key

from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("cosmograph")  # fetch version from package metadata
except PackageNotFoundError:
    __version__ = None



"""Base functionality of cosmograph"""

from typing import Dict, Any, Union, Callable, Sequence
from functools import cached_property, partial
from i2.doc_mint import inject_docstring_content

from cosmograph.widget import Cosmograph

from cosmograph.util import (
    CosmoKwargs,
    snake_to_camel_case,
    cosmograph_base_docs,
    cosmograph_base_signature,
    add_cosmo_param_descriptions,
)

from cosmograph.config import get_api_key

cosmo_base_sig = cosmograph_base_signature()
cosmo_base_params_doc_str = cosmograph_base_docs()


@inject_docstring_content(cosmo_base_params_doc_str, position=-1)
@cosmo_base_sig
def base_cosmo(**kwargs):
    """
    Thin layer over CosmographWidget to provide a base interface to the widget object.

    All this layer does is pass on the arguments to the CosmographWidget object.
    The only difference between this and the cosmograph object is that this one has
    an actual signature, with argument names, defaults, and type annotations.
    """
    return Cosmograph(**kwargs)


@add_cosmo_param_descriptions
def cosmo(
    data=None,
    *,
    # base cosmograph widget params ---------------------------------------------------
    points: object = None,
    links: object = None,
    point_x_by: str = None,
    point_y_by: str = None,
    point_size_by: str = None,
    point_color_by: str = None,
    point_color_palette: list[str] = None,
    point_color_by_map: Dict[str, Union[str, list[float]]] = None,
    point_color_strategy: str = None,
    point_label_by: str = None,
    point_color: Union[str, list[float]] = None,
    point_greyout_opacity: float = None,
    point_size: float = None,
    point_size_scale: float = None,
    point_sampling_distance: int = None,
    polygonal_selector_stroke_color: str = None,
    polygonal_selector_line_width: float = None,
    point_id_by: str = None,
    point_index_by: str = None,
    point_size_range: list[float] = None,
    point_size_strategy: str = None,
    point_label_weight_by: str = None,
    point_cluster_by: str = None,
    point_cluster_strength_by: str = None,
    cluster_positions_map: Dict[str, list[float]] = None,
    point_include_columns: list[str] = None,
    point_timeline_by: str = None,
    link_color: Union[str, list[float]] = None,
    link_greyout_opacity: float = None,
    link_width: float = None,
    link_width_scale: float = None,
    link_arrows: bool = None,
    link_arrows_size_scale: float = None,
    link_visibility_distance_range: list[float] = None,
    link_visibility_min_transparency: float = None,
    link_source_by: str = None,
    link_source_index_by: str = None,
    link_target_by: str = None,
    link_target_index_by: str = None,
    link_color_by: str = None,
    link_width_by: str = None,
    link_arrow_by: str = None,
    link_strength_by: str = None,
    link_strength_range: list[float] = None,
    link_include_columns: list[str] = None,
    link_timeline_by: str = None,
    disable_simulation: bool = None,
    simulation_decay: float = None,
    simulation_gravity: float = None,
    simulation_center: float = None,
    simulation_repulsion: float = None,
    simulation_repulsion_theta: float = None,
    simulation_repulsion_quadtree_levels: float = None,
    simulation_link_spring: float = None,
    simulation_link_distance: float = None,
    simulation_link_dist_random_variation_range: list[Any] = None,
    simulation_repulsion_from_mouse: float = None,
    simulation_friction: float = None,
    simulation_cluster_strength: float = None,
    background_color: Union[str, list[float]] = None,
    space_size: int = None,
    hovered_point_cursor: str = None,
    render_hovered_point_ring: bool = None,
    hovered_point_ring_color: Union[str, list[float]] = None,
    focused_point_ring_color: Union[str, list[float]] = None,
    focused_point_index: int = None,
    render_links: bool = None,
    curved_links: bool = None,
    curved_link_segments: int = None,
    curved_link_weight: float = None,
    curved_link_control_point_distance: float = None,
    use_quadtree: bool = None,
    show_fps_monitor: bool = None,
    pixel_ratio: float = None,
    # Set to True by default until it becomes the default in Cosmograph library
    scale_points_on_zoom: bool = True,
    scale_links_on_zoom: bool = None,
    initial_zoom_level: float = None,
    disable_zoom: bool = None,
    enable_drag: bool = None,
    fit_view_on_init: bool = None,
    fit_view_delay: float = None,
    fit_view_padding: float = None,
    fit_view_duration: float = None,
    fit_view_by_points_in_rect: list[list[float]] = None,
    random_seed: Union[int, str] = None,
    show_labels: bool = None,
    show_dynamic_labels: bool = None,
    show_labels_for: list[str] = None,
    show_top_labels: bool = None,
    show_top_labels_limit: int = None,
    static_label_weight: float = None,
    dynamic_label_weight: float = None,
    label_margin: float = None,
    show_hovered_point_label: bool = None,
    show_cluster_labels: bool = None,
    point_label_color: str = None,
    point_label_font_size: float = None,
    cluster_label_font_size: float = None,
    scale_cluster_labels: bool = None,
    label_padding: list[float] = None,
    use_point_color_strategy_for_cluster_labels: bool = None,
    select_cluster_on_label_click: bool = None,
    select_point_on_click: Union[bool, str] = None,
    select_point_on_label_click: Union[bool, str] = None,
    focus_point_on_click: bool = None,
    focus_point_on_label_click: bool = None,
    components_display_state_mode: Union[str, bool] = None,
    status_indicator_mode: Union[str, bool] = None,
    preserve_point_positions_on_data_update: bool = None,
    disable_point_size_legend: bool = None,
    disable_link_width_legend: bool = None,
    disable_point_color_legend: bool = None,
    disable_link_color_legend: bool = None,
    api_key: str = None,
    clicked_point_index: int = None,
    clicked_point_id: str = None,
    selected_point_indices: list[int] = None,
    selected_point_ids: list[str] = None,
):
    """
    Create an interactive network visualization with Cosmograph.

    Args:
        data: Convenience argument whose value will be used for either the points
    """
    # Get the arguments as a dictionary
    kwargs = locals().copy()

    # Remove data from kwargs as we'll handle it separately
    data = kwargs.pop("data")

    # Handle data assignment to points or links
    kwargs = prioritize_points(kwargs, data)

    # Remove None values
    kwargs = remove_none_values(kwargs)

    # Validate kwargs
    kwargs = validate_kwargs(kwargs)

    # If api_key is None, use the global one if available
    if kwargs.get('api_key') is None:
        global_api_key = get_api_key()
        if global_api_key is not None:
            kwargs['api_key'] = global_api_key

    # Make a Cosmograph widget instance with these kwargs
    return Cosmograph(**kwargs)


def process_cosmo_input(kwargs: CosmoKwargs) -> CosmoKwargs:
    data = kwargs.pop("data", None)
    return CosmoArguments(data, kwargs).prepare_kwargs()


def copy_points_and_links(kwargs):
    if kwargs.get("points", None) is not None:
        kwargs["points"] = kwargs["points"].copy()
    if kwargs.get("links", None) is not None:
        kwargs["links"] = kwargs["links"].copy()
    return kwargs


def remove_none_values(d):
    return {k: v for k, v in d.items() if v is not None}


def first_element(seq):
    return next(iter(seq), None)


def prioritize_points(kwargs, data=None):
    """
    A data resolver: Will priorities assigning data to points.

    The rules:
    - If `data` is not given (i.e. `None`), then `points` and `links` are returned as is.
    - If `data` is given, then:
        - If `points` is not given, then `points` is taken to be `data`.
        - If `links` is not given, then `links` is taken to be `data`.
        - If both `points` and `links` are given, raise an error.
    """
    if data is not None:
        if kwargs.get("points", None) is None:
            # If points is None, take data to be points (whether links is None or not)
            kwargs["points"] = data
        elif kwargs.get("links", None) is None:  # and points is not None
            # If points are given, but links is None, use data as links
            kwargs["links"] = data
        else:
            # If both points and links are given, raise an error
            raise ValueError(
                "Cannot specify `points` and `links` if you've specified `data`"
            )
    return kwargs


def validate_kwargs(kwargs):
    valid_names = set(cosmo_base_sig.names)
    invalid_keywords = kwargs.keys() - valid_names

    # check that all keys of kwargs are in valid_names
    if invalid_keywords:
        error_msg = "I didn't recognize some of your arguments:\n"

        if invalid_kws_that_are_aliases := invalid_keywords & set(argument_aliases):
            # make a message for the aliases, mentioning the keyword they are aliases for
            # and should be replaced by
            error_msg += "\n".join(
                f"You said `{kw}`: Did you mean `{argument_aliases[kw]}`?"
                for kw in invalid_kws_that_are_aliases
            )

        if invalid_kws_that_are_not_aliases := invalid_keywords - set(argument_aliases):
            t = ", ".join(invalid_kws_that_are_not_aliases)
            error_msg += f"And these I have no idea about: {t}"

        raise ValueError(error_msg)

    return kwargs


CosmoKwargsTrans = Callable[[CosmoKwargs], CosmoKwargs]
Data = Any
dflt_ingress = (copy_points_and_links,)


def __extra_cosmo_params(
    data=None,
    *,
    ingress: Sequence[CosmoKwargsTrans] = (copy_points_and_links,),
    copy_before_ingress: bool = True,
    data_resolution: Callable[[CosmoKwargs, Data], CosmoKwargs] = prioritize_points,
    validate_kwargs: Callable[[CosmoKwargs], CosmoKwargs] = validate_kwargs,
):
    """Just to hold the signature of the extra cosmo params"""


# @inject_docstring_content(cosmo_base_params_doc_str, position=-1)
# @cosmo_base_sig.inject_into_keyword_variadic
@inject_docstring_content(cosmo_base_params_doc_str, position=-1)
def cosmo_advanced(
    data=None,
    *,
    ingress: Sequence[CosmoKwargsTrans] = (),
    # base cosmograph widget params ---------------------------------------------------
    points: object = None,
    links: object = None,
    point_x_by: str = None,
    point_y_by: str = None,
    point_size_by: str = None,
    point_color_by: str = None,
    point_color_palette: list[str] = None,
    point_color_by_map: Dict[str, Union[str, list[float]]] = None,
    point_color_strategy: str = None,
    point_label_by: str = None,
    point_color: Union[str, list[float]] = None,
    point_greyout_opacity: float = None,
    point_size: float = None,
    point_size_scale: float = None,
    point_sampling_distance: int = None,
    polygonal_selector_stroke_color: str = None,
    polygonal_selector_line_width: float = None,
    point_id_by: str = None,
    point_index_by: str = None,
    point_size_range: list[float] = None,
    point_size_strategy: str = None,
    point_label_weight_by: str = None,
    point_cluster_by: str = None,
    point_cluster_strength_by: str = None,
    cluster_positions_map: Dict[str, list[float]] = None,
    point_include_columns: list[str] = None,
    point_timeline_by: str = None,
    link_color: Union[str, list[float]] = None,
    link_greyout_opacity: float = None,
    link_width: float = None,
    link_width_scale: float = None,
    link_arrows: bool = None,
    link_arrows_size_scale: float = None,
    link_visibility_distance_range: list[float] = None,
    link_visibility_min_transparency: float = None,
    link_source_by: str = None,
    link_source_index_by: str = None,
    link_target_by: str = None,
    link_target_index_by: str = None,
    link_color_by: str = None,
    link_width_by: str = None,
    link_arrow_by: str = None,
    link_strength_by: str = None,
    link_strength_range: list[float] = None,
    link_include_columns: list[str] = None,
    link_timeline_by: str = None,
    disable_simulation: bool = None,
    simulation_decay: float = None,
    simulation_gravity: float = None,
    simulation_center: float = None,
    simulation_repulsion: float = None,
    simulation_repulsion_theta: float = None,
    simulation_repulsion_quadtree_levels: float = None,
    simulation_link_spring: float = None,
    simulation_link_distance: float = None,
    simulation_link_dist_random_variation_range: list[Any] = None,
    simulation_repulsion_from_mouse: float = None,
    simulation_friction: float = None,
    simulation_cluster_strength: float = None,
    background_color: Union[str, list[float]] = None,
    space_size: int = None,
    hovered_point_cursor: str = None,
    render_hovered_point_ring: bool = None,
    hovered_point_ring_color: Union[str, list[float]] = None,
    focused_point_ring_color: Union[str, list[float]] = None,
    focused_point_index: int = None,
    show_focused_point_label: bool = None,
    render_links: bool = None,
    curved_links: bool = None,
    curved_link_segments: int = None,
    curved_link_weight: float = None,
    curved_link_control_point_distance: float = None,
    use_quadtree: bool = None,
    show_FPS_monitor: bool = None,
    pixel_ratio: float = None,
    # Set to True by default until it becomes the default in Cosmograph library
    scale_points_on_zoom: bool = True,
    scale_links_on_zoom: bool = None,
    initial_zoom_level: float = None,
    disable_zoom: bool = None,
    enable_drag: bool = None,
    fit_view_on_init: bool = None,
    fit_view_delay: float = None,
    fit_view_padding: float = None,
    fit_view_duration: float = None,
    fit_view_by_points_in_rect: list[list[float]] = None,
    random_seed: Union[int, str] = None,
    show_labels: bool = None,
    show_dynamic_labels: bool = None,
    show_labels_for: list[str] = None,
    show_top_labels: bool = None,
    show_top_labels_limit: int = None,
    static_label_weight: float = None,
    dynamic_label_weight: float = None,
    label_margin: float = None,
    show_hovered_point_label: bool = None,
    show_cluster_labels: bool = None,
    point_label_color: str = None,
    point_label_font_size: float = None,
    cluster_label_font_size: float = None,
    scale_cluster_labels: bool = None,
    label_padding: list[float] = None,
    use_point_color_strategy_for_cluster_labels: bool = None,
    select_cluster_on_label_click: bool = None,
    select_point_on_click: Union[bool, str] = None,
    select_point_on_label_click: Union[bool, str] = None,
    focus_point_on_click: bool = None,
    focus_point_on_label_click: bool = None,
    components_display_state_mode: Union[str, bool] = None,
    status_indicator_mode: Union[str, bool] = None,
    preserve_point_positions_on_data_update: bool = None,
    disable_point_size_legend: bool = None,
    disable_link_width_legend: bool = None,
    disable_point_color_legend: bool = None,
    disable_link_color_legend: bool = None,
    api_key: str = None,
    clicked_point_index: int = None,
    clicked_point_id: str = None,
    selected_point_indices: list[int] = None,
    selected_point_ids: list[str] = None,
    changePoints: Callable[[Dict[str, Any]], Any] = None,
    changeLinks: Callable[[Dict[str, Any]], Any] = None,
    # extra params ---------------------------------------------------------------------
    copy_before_ingress: bool = True,  # whether to make a copy the points and links before applying ingress
    data_resolution: Callable[
        [CosmoKwargs, Data], CosmoKwargs
    ] = prioritize_points,  # What to do with data (how to integrate it into the cosmo kwargs)
    validate_kwargs: Callable[
        [CosmoKwargs], CosmoKwargs
    ] = validate_kwargs,  # function to apply after the ingress transformations
):
    """
    Thin layer over CosmographWidget to provide a base interface to the widget object,
    with signature, docs with argument descriptions, and a more flexible interface
    comprising some data processing, argument aliasing and validation, error handling,
    etc.

    .. deprecated:: 1.0.0
        This function is deprecated. Use the simpler `cosmo` function instead.

    :param data: Convenience argument whose value will be used for either the points
        or links argument.
    :param ingress: A function or sequence of functions that will be applied to the
        kwargs before they are passed to the Cosmograph constructor.
    :param copy_before_ingress: Whether to make a copy of the points and links before
        applying the ingress transformations. This is so that the original data is not
        modified by the ingress functions.
    :param data_resolution: A function that will be used to resolve the data argument.
        This is the function that takes care of how to handle the data argument,
    :param validate_kwargs: A function that will be applied to the kwargs after the
        ingress transformations. This is the function that will validate the kwargs
        before they are passed to the Cosmograph constructor.

    """
    import warnings
    warnings.warn(
        "cosmo_advanced is deprecated and will be removed in a future version. "
        "Use the simpler 'cosmo' function instead.",
        DeprecationWarning,
        stacklevel=2
    )

    # Get the arguments as a dictionary
    kwargs = locals().copy()

    # If there's no data arguments (data, points, links)
    # return a partial function with other settings
    if all(val is None for val in map(kwargs.get, ["data", "points", "links"])):
        _ = kwargs.pop("data")
        return partial(cosmo, **kwargs)

    # extract all arguments that are not base cosmograph arguments
    # (and leaving kwargs with only the cosmograph arguments)
    data, ingress, copy_before_ingress, data_resolution, validate_kwargs = map(
        kwargs.pop,
        [
            "data",
            "ingress",
            "copy_before_ingress",
            "data_resolution",
            "validate_kwargs",
        ],
    )

    # put data in points or links according to the
    kwargs = data_resolution(kwargs, data)

    # process ingress
    if callable(ingress):
        ingress = (ingress,)

    if copy_before_ingress:
        ingress = (copy_points_and_links,) + tuple(ingress)

    # apply the ingress transformations one by one
    for kwargs_trans in ingress:
        kwargs = kwargs_trans(kwargs)

    # remove all items that have None values
    kwargs = remove_none_values(kwargs)

    # validate the kwargs
    kwargs = validate_kwargs(kwargs)

    # If api_key is None, use the global one if available
    if kwargs.get('api_key') is None:
        global_api_key = get_api_key()
        if global_api_key is not None:
            kwargs['api_key'] = global_api_key

    # Make a Cosmograph widget instance with these kwargs
    return Cosmograph(**kwargs)


# cosmo.ingress = dflt_cosmo_ingress


# Note: Possibly deprecated
class CosmoArguments:
    """
    Container for cosmograph arguments, along with diagnosis and processing methods
    """

    def __init__(self, data=None, kwargs=()):
        self.data = data
        # TODO: Should any Nones actually be passed on?
        self.kwargs = dict(kwargs)

    def prepare_kwargs(self):
        self.validate_kwargs()
        points, links = self.points_and_links()
        kwargs = remove_none_values(dict(self.kwargs, points=points, links=links))
        return kwargs

    def validate_kwargs(self):
        validate_kwargs(self.kwargs)

    def points_and_links(self):
        """
        Get the points and links from the arguments, and raise an error if both are
        specified and `data` is also specified.

        The rules:
        - If `data` is not given (i.e. `None`), then `points` and `links` are returned as is.
        - If `data` is given, then:
            - If `points` is not given, then `points` is taken to be `data`.
            - If `links` is not given, then `links` is taken to be `data`.
            - If both `points` and `links` are given, raise an error.
        """
        data = self.data
        points, links = self.get(["points", "links"], default=None)

        if data is not None:
            if points is None:
                # If points is None, take data to be points (whether links is None or not)
                points = data
            elif links is None:  # and points is not None
                # If points are given, but links is None, use data as links
                links = data
            else:
                # If both points and links are given, raise an error
                raise ValueError(
                    "Cannot specify `points` and `links` if you've specified `data`"
                )

        return points, links

    @cached_property
    def non_none_names(self):
        return {k for k, v in self.kwargs.items() if v is not None}

    def get(self, name, default=None):
        if isinstance(name, (list, tuple)):
            return [self.kwargs.get(k, default) for k in name]
        return self.kwargs.get(name, default)

    def non_nones(self, names):
        return self.non_nones_names & set(names)

    def all_are_none(self, names):
        return self.non_nones(names) == set()

    def none_are_none(self, names):
        return len(self.non_nones(names)) == len(names)


# Make aliases for the arguments
argument_aliases = {
    "nodes": "points",
    "edges": "links",
}
# camel case aliases
argument_aliases = dict(
    argument_aliases, **{snake_to_camel_case(k): k for k in cosmo_base_sig.names}
)
argument_aliases = dict(
    argument_aliases,
    **{
        snake_to_camel_case(k, first_char_trans=str.upper): k
        for k in cosmo_base_sig.names
    },
)

# --------------------------------------------------------------------------------------
# Some old code that might be useful

from typing import List
from pydantic import BaseModel

NodeId = str  # TODO: Also include ints? Oblige ids to be valid python identifiers?


class Node(BaseModel):
    id: NodeId


class Link(BaseModel):
    source: NodeId
    target: NodeId


# TODO: Find a way to make these have a .validate (with pydantic)
Nodes = List[Node]
Links = List[Link]


class GraphJson(BaseModel):
    links: Links
    nodes: Nodes = []


from cosmograph.validation import ensure_json_string


html_code_data_def_template = """
<div>
    <canvas></canvas>
</div>

<script>
    const data = {json_data_str}
    const canvas = document.querySelector("canvas");
    const graph = new cosmos.Graph(canvas);
    graph.setData(data.nodes, data.links);
    graph.fitView();
</script>
"""


# TODO: Function should be moved to a better place
def mk_html_code(data):
    data = ensure_json_string(data)
    return html_code_data_def_template.format(json_data_str=data)



"""Utility functions for Cosmograph widget including memory caching and data conversion."""

import logging
from joblib import Memory
from pathlib import Path
import pyarrow as pa

# Setup joblib cache
CACHE_DIR = Path(__file__).parent / ".cosmograph_cache"
CACHE_DIR.mkdir(exist_ok=True)
memory = Memory(CACHE_DIR, verbose=0)  # Set verbose=1 for debugging cache misses

logger = logging.getLogger(__name__)


@memory.cache
def get_buffered_arrow_table(df):
  """Converts a Pandas DataFrame to a buffered Arrow IPC stream format.

  This function is cached using joblib.Memory.
  """

  if df is None:
      return None
  try:
      df_int32 = df.select_dtypes(include=["int64"]).astype("int32")
      df[df_int32.columns] = df_int32
      table = pa.Table.from_pandas(df)
      sink = pa.BufferOutputStream()
      with pa.ipc.new_stream(sink, table.schema) as writer:
          writer.write(table)
      buffer = sink.getvalue()
      return buffer.to_pybytes()
  except Exception as e:
      logger.warning("Failed to convert DataFrame to buffered Arrow table: %s", str(e))
      return None
